/*
  Hatari - main.c

  This file is distributed under the GNU Public License, version 2 or at
  your option any later version. Read the file gpl.txt for details.

  Main initialization and event handling routines.
*/

#include <time.h>
#include <signal.h>
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>

#include <SDL.h>

#include "main.h"
#include "configuration.h"
#include "dialog.h"
#include "audio.h"
#include "file.h"
#include "../m68000.h"
#include "hostcall.h"
#include "input.h"
#include "keymap.h"
#include "misc.h"
#include "screen.h"
#include "sdlgui.h"
#include "shortcut.h"


#define FORCE_WORKING_DIR                 /* Set default directory to cwd */


BOOL bQuitProgram=FALSE;                  /* Flag to quit program cleanly */
BOOL bUseFullscreen=FALSE;
BOOL bEmulationActive=TRUE;               /* Run emulation when started */
BOOL bAppActive = FALSE;
char szBootDiscImage[MAX_FILENAME_LENGTH] = { "" };

char szWorkingDir[MAX_FILENAME_LENGTH] = { "" };
char szCurrentDir[MAX_FILENAME_LENGTH] = { "" };

int ratonx=0,ratony=0,incraton=2;
int u=0,d=0,l=0,r=0,ul=0,ur=0,dl=0,dr=0,raton=0;

typedef struct {
    uint16_t               pressure;
    uint16_t              x;
    uint16_t               y;
    uint16_t               pad;
    struct timeval  stamp;
} TS_EVENT;

TS_EVENT ts_event;
int fd;						// Buffer pa lectura de tactil
static int tactilx,tactily;
char autocalib=1;
static int calibX=0, calibY=0;

/*-----------------------------------------------------------------------*/
/*
  Error handler
*/
void Main_SysError(char *Error,char *Title)
{
  fprintf(stderr,"%s : %s\n",Title,Error);
}


/*-----------------------------------------------------------------------*/
/*
  Bring up message(handles full-screen as well as Window)
*/
int Main_Message(char *lpText, char *lpCaption/*,unsigned int uType*/)
{
  int Ret=0;

  /* Show message */
  fprintf(stderr,"%s: %s\n", lpCaption, lpText);

  return(Ret);
}


/*-----------------------------------------------------------------------*/
/*
  Pause emulation, stop sound
*/
void Main_PauseEmulation(void)
{
  if( bEmulationActive )
  {
    Audio_EnableAudio(FALSE);
    bEmulationActive = FALSE;
  }
}

/*-----------------------------------------------------------------------*/
/*
  Start emulation
*/
void Main_UnPauseEmulation(void)
{
  if( !bEmulationActive )
  {
    Audio_EnableAudio(ConfigureParams.Sound.bEnableSound);
    bFullScreenHold = FALSE;      /* Release hold  */

    bEmulationActive = TRUE;
  }
}

/* ----------------------------------------------------------------------- */
/*
  Message handler
  Here we process the SDL events (keyboard, mouse, ...) and map it to
  Atari IKBD events.
*/
void Main_EventHandler()
{
  SDL_Event event;

  if (raton){

  if (r){
      ratonx += incraton;
      input.motion_x +=incraton;
      input.abs_x =ratonx;
    }

  if (l){
      ratonx -= incraton;
      input.motion_x -=incraton;
      input.abs_x =ratonx;
    }

  if (u){
      ratony -= incraton;
      input.motion_y -=incraton;
      input.abs_y =ratony;
    }

   if (d){
       ratony += incraton;
       input.motion_y +=incraton;
       input.abs_y =ratony;
     }


   if (ul){
       ratony -= incraton;
       ratonx -= incraton;
       input.motion_y -=incraton;
       input.motion_x -=incraton;
       input.abs_x =ratonx;
       input.abs_y =ratony;
     }

   if (ur){
       ratony -= incraton;
       ratonx += incraton;
       input.motion_y -=incraton;
       input.motion_x +=incraton;
       input.abs_x =ratonx;
       input.abs_y =ratony;
     }

  if (dl){
      ratony += incraton;
      ratonx -= incraton;
      input.motion_y +=incraton;
      input.motion_x -=incraton;
      input.abs_x =ratonx;
      input.abs_y =ratony;
    }

  if (dr){
      ratony += incraton;
      ratonx += incraton;
      input.motion_y +=incraton;
      input.motion_x +=incraton;
      input.abs_x =ratonx;
      input.abs_y =ratony;
    }
  }

  if (r)
    if (ratonx<320){
      ratonx += incraton;
      input.motion_x +=incraton;
      input.abs_x =ratonx;
    }

  if (l)
    if (ratonx>0){
      ratonx -= incraton;
      input.motion_x -=incraton;
      input.abs_x =ratonx;
    }

  if (u)
    if (ratony>0){
      ratony -= incraton;
      input.motion_y -=incraton;
      input.abs_y =ratony;
    }

   if (d)
     if (ratony<200){
       ratony += incraton;
       input.motion_y +=incraton;
       input.abs_y =ratony;
     }


   if (ul)
     if ((ratonx>0) || (ratony>0)){
       ratony -= incraton;
       ratonx -= incraton;
       input.motion_y -=incraton;
       input.motion_x -=incraton;
       input.abs_x =ratonx;
       input.abs_y =ratony;
     }

   if (ur)
     if ((ratonx<320) || (ratony>0)){
       ratony -= incraton;
       ratonx += incraton;
       input.motion_y -=incraton;
       input.motion_x +=incraton;
       input.abs_x =ratonx;
       input.abs_y =ratony;
     }

  if (dl)
    if ((ratonx>0) || (ratony<200)){
      ratony += incraton;
      ratonx -= incraton;
      input.motion_y +=incraton;
      input.motion_x -=incraton;
      input.abs_x =ratonx;
      input.abs_y =ratony;
    }

  if (dr)
    if ((ratonx<320) || (ratony<200)){
      ratony += incraton;
      ratonx += incraton;
      input.motion_y +=incraton;
      input.motion_x +=incraton;
      input.abs_x =ratonx;
      input.abs_y =ratony;
    }



// Nuevo codigo de raton


	read(fd, &ts_event, sizeof(TS_EVENT));

//	autocalib=1;
	calibX=0;
	calibY=0;

	if ((autocalib==1)&&(ts_event.pressure==0)){
	        calibX = ((ts_event.x-200)*320/3750)/4;
                calibY = (((ts_event.y-200)*240/3750))/4;
                autocalib=0;
        }


	if (ts_event.pressure>0){

		tactilx = ((ts_event.x-200)*320/3750)-calibX;
		tactily = (240 - ((ts_event.y-200)*240/3750))-calibY;


		if (tactily>200)
			tactily=200;

		if (tactily<0)
			tactily=0;

		if (tactilx>320)
			tactilx=320;

		if (tactilx<0)
			tactilx=0;
		

		ratonx=tactilx;
		ratony=tactily;

		input.abs_x =ratonx;
		input.abs_y =ratony;

                Input_MousePress (SDL_BUTTON_LEFT);

	}else
		Input_MouseRelease (SDL_BUTTON_LEFT);
	


// Nuevo codigo de raton



  while (mySDL_PollEvent (&event))
   switch( event.type )
   {
    case SDL_QUIT:
       bQuitProgram = TRUE;
       SDL_Quit ();
       exit (0);
       break;
    case SDL_MOUSEMOTION:               /* Read/Update internal mouse position */
       input.motion_x += event.motion.xrel;
       input.motion_y += event.motion.yrel;
       input.abs_x = event.motion.x;
       input.abs_y = event.motion.y;
       break;
    case SDL_MOUSEBUTTONDOWN:
       Input_MousePress (event.button.button);
       break;
    case SDL_MOUSEBUTTONUP:
       Input_MouseRelease (event.button.button);
       break;
/*    case SDL_KEYDOWN:
       Keymap_KeyDown(&event.key.keysym);
       break;
    case SDL_KEYUP:
       Keymap_KeyUp(&event.key.keysym);
       break;
*/

    case SDL_KEYDOWN:
      switch( event.key.keysym.sym )
      {
	case SDLK_e:
	  bQuitProgram = TRUE;
	  SDL_Quit ();
	  exit (0);
	  break;
	case SDLK_PAUSE:
	  event.key.keysym.sym=SDLK_ESCAPE;
	  Keymap_KeyDown(&event.key.keysym);
	  break;
	case SDLK_c:
	  event.key.keysym.sym=SDLK_F1;
	  Keymap_KeyDown(&event.key.keysym);
	  break;
	case SDLK_d:
	  event.key.keysym.sym=SDLK_F3;
	  Keymap_KeyDown(&event.key.keysym);
	  break;
	case SDLK_b:
	  event.key.keysym.sym=SDLK_RSHIFT;
	  Keymap_KeyDown(&event.key.keysym);
          break;
           case SDLK_RIGHT:
             r=1;
             event.key.keysym.sym=SDLK_RIGHT;
             Keymap_KeyDown(&event.key.keysym);
	     break;
           case SDLK_LEFT:
             l=1;
             event.key.keysym.sym=SDLK_LEFT;
             Keymap_KeyDown(&event.key.keysym);
             break;
           case SDLK_UP:
             u=1;
             event.key.keysym.sym=SDLK_UP;
             Keymap_KeyDown(&event.key.keysym);
             break;
           case SDLK_DOWN:
             d=1;
             event.key.keysym.sym=SDLK_DOWN;
             Keymap_KeyDown(&event.key.keysym); 
	     break;
           case SDLK_MINUS:
             Input_MousePress (SDL_BUTTON_LEFT);
             break;
           case SDLK_PLUS:
             Input_MousePress (SDL_BUTTON_RIGHT);
             raton=1;
             break;
       }
    Keymap_KeyDown(&event.key.keysym);
    break;
    case SDL_KEYUP:
      switch( event.key.keysym.sym )
      {
	case SDLK_PAUSE:
	  event.key.keysym.sym=SDLK_ESCAPE;
	  Keymap_KeyUp(&event.key.keysym);
	  break;
	case SDLK_c:
	  event.key.keysym.sym=SDLK_F1;
	  Keymap_KeyUp(&event.key.keysym);
	  break;
	case SDLK_d:
	  event.key.keysym.sym=SDLK_F3;
	  Keymap_KeyUp(&event.key.keysym);
	  break;
	case SDLK_b:
	  event.key.keysym.sym=SDLK_RSHIFT;
	  Keymap_KeyUp(&event.key.keysym);
          break;
           case SDLK_RIGHT:
             r=0;
             event.key.keysym.sym=SDLK_RIGHT;
             Keymap_KeyUp(&event.key.keysym);
	     break;
           case SDLK_LEFT:
             l=0;
             event.key.keysym.sym=SDLK_LEFT;
             Keymap_KeyUp(&event.key.keysym);
             break;
           case SDLK_UP:
             u=0;
             event.key.keysym.sym=SDLK_UP;
             Keymap_KeyUp(&event.key.keysym);
             break;
           case SDLK_DOWN:
             d=0;
             event.key.keysym.sym=SDLK_DOWN;
             Keymap_KeyUp(&event.key.keysym);
	     break;
           case SDLK_PLUS:
             Input_MouseRelease (SDL_BUTTON_RIGHT);
             raton=0;
             if (ratonx>320)
               ratonx=320;

             if (ratonx<0)
               ratonx=0;

             if (ratony>200)
               ratony=200;
 
             if (ratony<0)
               ratony=0;
             break;
           case SDLK_MINUS:
             Input_MouseRelease (SDL_BUTTON_LEFT);
             break; 
      }

    Keymap_KeyUp(&event.key.keysym);
    break;
   }
  Input_Update ();
}


/*-----------------------------------------------------------------------*/
/*
  Check for any passed parameters
*/
void Main_ReadParameters(int argc, char *argv[])
{
  int i;

  /* Scan for any which we can use */
  for(i=1; i<argc; i++)
  {
    if (strlen(argv[i])>0)
    {
      if (!strcmp(argv[i],"--help") || !strcmp(argv[i],"-h"))
      {
        printf("Usage:\n frontier [options]\n"
               "Where options are:\n"
               "  --help or -h          Print this help text and exit.\n"
               "  --fullscreen or -f    Try to use fullscreen mode.\n"
               "  --nosound             Disable sound (faster!).\n"
              );
        exit(0);
      }
      else if (!strcmp(argv[i],"--fullscreen") || !strcmp(argv[i],"-f"))
      {
        bUseFullscreen=TRUE;
      }
      else if ( !strcmp(argv[i],"--nosound") )
      {
        bDisableSound=TRUE;
        ConfigureParams.Sound.bEnableSound = FALSE;
      }
      else
      {
	      /* some time make it possible to read alternative
	       * names for fe2.bin from command line */
	      fprintf(stderr,"Illegal parameter: %s\n",argv[i]);
      }
    }
  }
}


/*-----------------------------------------------------------------------*/
/*
  Initialise emulation
*/
void Main_Init(void)
{
  /* Init SDL's video subsystem. Note: Audio and joystick subsystems
     will be initialized later (failures there are not fatal). */
  if(SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER) < 0)
  {
    fprintf(stderr, "Could not initialize the SDL library:\n %s\n", SDL_GetError() );
    exit(-1);
  }
  
  Misc_SeedRandom(1043618);
  SDLGui_Init();
  Screen_Init();
  Init680x0();                  /* Init CPU emulation */
  Audio_Init();
  Keymap_Init();

  GemDOS_Init();
  GemDOS_InitDrives();

  if(bQuitProgram)
  {
    SDL_Quit();
    exit(-2);
  }
}


/*-----------------------------------------------------------------------*/
/*
  Un-Initialise emulation
*/
void Main_UnInit(void)
{
  Screen_ReturnFromFullScreen();
  GemDOS_UnInitDrives();
  Audio_UnInit();
  SDLGui_UnInit();
  Screen_UnInit();

  /* SDL uninit: */
  SDL_Quit();
}

static Uint32 vbl_callback ()
{
	FlagException (0);
	return 21;
}

void sig_handler (int signum)
{
	if (signum == SIGSEGV) {
		printf ("Segfault! All is lost! Abandon ship!\n");
		Call_DumpDebug ();
		abort ();
	}
}

/*-----------------------------------------------------------------------*/
/*
  Main
*/
int main(int argc, char *argv[])
{
  struct sigaction act;

  act.sa_handler = sig_handler;
  act.sa_flags = 0;
  sigaction (SIGSEGV, &act, NULL);
	
  /* Generate random seed */
  srand( time(NULL) );

  /* Get working directory, if in MSDev force */
  Misc_FindWorkingDirectory(argv[0]);
#ifdef FORCE_WORKING_DIR
  getcwd(szWorkingDir, MAX_FILENAME_LENGTH);
#endif

  /* Set default configuration values: */
  Configuration_SetDefault();

  /* Check for any passed parameters */
  Main_ReadParameters(argc, argv);

  fd = open( "/dev/touchscreen/wm97xx", O_RDONLY | O_NOCTTY );	// Lectura del tactil

  if (fd){
 
  /* Init emulator system */
  Main_Init();

  /* Switch immediately to fullscreen if user wants to */
  if( bUseFullscreen )
    Screen_EnterFullScreen();

  SDL_AddTimer (22, &vbl_callback, NULL);
  
  /* Run emulation */
  Main_UnPauseEmulation();
  Start680x0();                 /* Start emulation */

  /* Un-init emulation system */
  Main_UnInit();
 }
  return(0);
}

int mySDL_PollEvent(SDL_Event *event)  {
    int _r= SDL_PollEvent(event);
    if (!_r)
        return 0;
    if (event->type == SDL_ACTIVEEVENT) {
        if (event->active.state == SDL_APPINPUTFOCUS && !event->active.gain) {
            suspendAudio();
            for (;;) {
                _r = SDL_WaitEvent(event);
                if (!_r)
                    continue;
                if (event->type == SDL_QUIT)
                    return 1;
                if (event->type != SDL_ACTIVEEVENT)
                    continue;
                if (event->active.state == SDL_APPINPUTFOCUS && event->active.gain) {
                    resumeAudio();
                        return 1;
                }
            }
        }
    }
  return _r;
}

// if audio use sdl-mixer-lib
void suspendAudio()
{
  Mix_SuspendAudio();
}

int resumeAudio() {
  int r = Mix_ResumeAudio();
  if(r == -1){
    closeAudio();
    return -1;
  }
}

void closeAudio(void) {
      Mix_CloseAudio();
}


